'' =================================================================================================
''
''   File....... uSDX4FileIOMaster
''   Device..... uSD X4 Duplicator (1x Master, 3x Clients)
''   Purpose.... Main Controller for Master & 3 Client uSD Socket
''
''               Design Methodology
''               ==================
''               
''      1 x Clusters = Number of Sectors                               
''      1 x Sector = 512 bytes (128 longs)                
''               
''   Author..... Kenichi Kato (R & D)
''               Board Design under Copyright (c) 2011 Katotec Pte. Ltd.
''               (All Source Contains Within Are Under MIT License, Please See Below For Terms Of Use)
''   E-mail..... kkato@katotechnology.com
''   Started.... 22 Dec 2011
''   Updated....
''              28 Mar 2012: Added Start() & Stop() 
''              29 Mar 2012: Tested Sector Read/Write
''              30 Mar 2012: Kye's read/writeLowLevelSector() solved the sector problem
'' =================================================================================================
CON

  '--- --- --- --- --- --- --- --- --- ---  
  '---          Disk Setting (should not change)
  '--- --- --- --- --- --- --- --- --- ---
  _bytesPerSector = 512
  _longsPerSector = 128
  _waitToFlush    = 30  'msec  
  

  '--- --- --- --- --- --- --- --- --- ---  
  '---          uSD (Master)
  '--- --- --- --- --- --- --- --- --- ---  
  _Mas_DO        = 0
  _Mas_Clk       = 1
  _Mas_DI        = 2
  _Mas_CS        = 3
  _Mas_CD        = -1
  _Mas_WP        = -1

  '--- --- --- --- --- --- --- --- --- ---  
  '---          uSD (Client 1)
  '--- --- --- --- --- --- --- --- --- ---  
  _So1_DO        = 4
  _So1_Clk       = 5
  _So1_DI        = 6
  _So1_CS        = 7
  _So1_CD        = -1
  _So1_WP        = -1

  '--- --- --- --- --- --- --- --- --- ---  
  '---          uSD (Client 2)
  '--- --- --- --- --- --- --- --- --- ---  
  _So2_DO        = 8
  _So2_Clk       = 9
  _So2_DI        = 10
  _So2_CS        = 11
  _So2_CD        = -1
  _So2_WP        = -1

  '--- --- --- --- --- --- --- --- --- ---  
  '---          uSD (Client 3)
  '--- --- --- --- --- --- --- --- --- ---  
  _So3_DO        = 12
  _So3_Clk       = 13
  _So3_DI        = 14
  _So3_CS        = 15
  _So3_CD        = -1
  _So3_WP        = -1


  '--- --- --- --- --- --- --- --- --- ---  
  '---          Status Flags Definition
  '--- --- --- --- --- --- --- --- --- ---  
  _masCrdDet    = $1
  _c1CrdDet     = $2
  _c2CrdDet     = $4
  _c3CrdDet     = $8

  _SecFileSet   = $200
  _modeSecFile  = $400
  _UserInt      = $1_000
  _OpsOrNot     = $10_000
  _OpsSuccess   = $20_000

  _errSecCopy   = $1_000_000

  _debugFlag2   = $40_000_000 
  _debugFlag1   = $80_000_000



OBJ

  'Local Controlled
  uSDMas    : "X4Mas_SD-MMC_FATEngine.spin"                     '(Cog x 4)
  uSDCli[3] : "X4Cli_SD-MMC_FATEngine.spin"                     '(Cog x 4)
  
  
  'Traffic controller will run in a Cog                 '(Cog x 1)

  '*** Debugging
  DBG     : "FullDuplexSerialExt.spin"  '(Cog x 1)
    


VAR

  long msecValue
  'long masTotalSectors
  long uSDX4ControllerStack[128]
  long totalSectors[4]   'NOTE: Used as storing actual Sector#, starting from 0 ~
  long readSectorBlockBuffer[_longsPerSector]
  long writeSectorBlockBuffer[_longsPerSector]
  

  byte uSDCogChecker[4]
  'byte uSDBuffer[512]
  byte sectorsPerCluster[4]
  byte uSDX4ControllerCogID
  
  

PUB Start(statusAddress, msValue)
{{
  Start will only load drivers for uSD socket with uSD card inserted.  
}}

  '------------------
  '[Var] - Init
  '------------------
  msecValue := msValue  'Used in Pause()
  long[statusAddress] ^= statusAddress & _debugFlag1  'Clear Bit for debugging use


  'Check if loaded before
  StopX4uSDController


  'Launch Controller
  'uSDX4ControllerCogID := cognew(LaunchX4uSDController(statusAddress, readSectorBlockBuffer, writeSectorBlockBuffer), @uSDX4ControllerStack) + 1
  uSDX4ControllerCogID := cognew(LaunchX4uSDController(statusAddress), @uSDX4ControllerStack) + 1

  return

    


PUB StopX4uSDController

  If(uSDX4ControllerCogID)
    cogstop(uSDX4ControllerCogID~ -1)

  return


 
PUB LaunchX4uSDController(statusAddress) | curSector, i, errorSectorMismatch 
{{
  Main controller for back-end read/write operations
}}

  '*** *** *** *** *** *** 
  '*** Debugging Use Only
  '*** *** *** *** *** *** 
  DBG.Start(31, 30, 0, 115_200)
  Pause(500)
  DBG.Tx(0)
  DBG.Str(String(13, "======================================"))
  DBG.Str(String(13, "In LaunchX4uSDControllerCog Debug Mode")) 
  DBG.Str(String(13, "======================================", 13, 13))
  Pause(500)
  '*** *** *** *** *** *** 

   

  '------------------
  '[Var] - Init
  '------------------
  curSector := 0
  errorSectorMismatch := false
  i := 0
  repeat i from 0 to 3
    totalSectors[i] := 0
    uSDCogChecker[i] := 0
    sectorsPerCluster[i] := 0
  'longfill(@uSDBuffer, 0, _longsPerSector)  'Cleared block buffer
'  longfill(readStoreBlock, 0, _longsPerSector)
'  longfill(writeStoreBlock, 0, _longsPerSector)
  longfill(@readSectorBlockBuffer, 0 , _longsPerSector)
  longfill(@writeSectorBlockBuffer, 0 , _longsPerSector)
  

{
  '*** *** *** *** *** *** 
  '*** Debugging Use Only
  '*** *** *** *** *** ***
  'Test accessing memory
  repeat i from 0 to 127
    DBG.Hex(readSectorBlockBuffer[i], 8)
    DBG.Str(String(","))
  DBG.Str(String(13,13))
  'Store dummy data
  repeat i from 0 to 127
    readSectorBlockBuffer[i] := cnt
  repeat i from 0 to 127
    DBG.Hex(readSectorBlockBuffer[i], 8)
    DBG.Str(String(","))
  '*** *** *** *** *** ***
}    
    
      



  '------------------
  '[OBJ] - Just make sure Cogs not loaded
  '------------------
  Stop(statusAddress)

   
   
  '------------------
  '[OBJ] - Load Activated Drivers
  '------------------
  if(long[statusAddress] & _masCrdDet)
    'Master inserted, it should
    'Need to put Error capture in the future
    uSDMas.FATEngineStart(_Mas_DO, _Mas_Clk, _Mas_DI, _Mas_CS, _Mas_WP, _Mas_CD, -1, -1, -1)
    uSDMas.mountPartition(0)
    uSDCogChecker[0] := 1
    
  if(long[statusAddress] & _c1CrdDet)
    'Client 1
    uSDCli[0].FATEngineStart(_So1_DO, _So1_Clk, _So1_DI, _So1_CS, _So1_WP, _So1_CD, -1, -1, -1)
    uSDCli[0].mountPartition(0)
    uSDCogChecker[1] := 1

  if(long[statusAddress] & _c2CrdDet)
    'Client 2
    uSDCli[1].FATEngineStart(_So2_DO, _So2_Clk, _So2_DI, _So2_CS, _So2_WP, _So1_CD, -1, -1, -1)
    uSDCli[1].mountPartition(0)
    uSDCogChecker[2] := 1

  if(long[statusAddress] & _c3CrdDet)
    'Client 3
    uSDCli[2].FATEngineStart(_So3_DO, _So3_Clk, _So3_DI, _So3_CS, _So3_WP, _So3_CD, -1, -1, -1)
    uSDCli[2].mountPartition(0)
    uSDCogChecker[3] := 1


  '***
  '*** Debugging User Only
  '***
  long[statusAddress] |= _debugFlag1   'Set Bit  - PASSED
  

  Pause(100)


  '------------------
  '[APP] - Sanity-check before proceeding 
  '------------------
  if(long[statusAddress] & _SecFileSet)
    'OK, Set & good-to-go 


    '------------------
    '[APP] - Run Copy Operation (base on Mode Selection)
    '------------------
    if(long[statusAddress] & _modeSecFile)
    
      '@@@ @@@ @@@ @@@ @@@
      '@@@ SECTOR COPY @@@  
      '@@@ @@@ @@@ @@@ @@@
      'The Client card(s) must have the same number of Sectors per Cluster as the Master card
      'Operation will stop if different

{
      '[Master] Read Number of Clusters & Total Sectors
      '-------- 
      sectorsPerCluster[0] := uSD[0].partitionSectorsPerCluster 
      totalSectors[0] := (sectorsPerCluster[0] * uSD[0].partitionCountOfClusters) - 1  'Setting it as Sector#


      '[Client 1 to 3] For those mounted,
      'Read Total Number of Sectors per Disk

      repeat i from 1 to 3
        if(uSDCogChecker[i])
          totalSectors[i] := uSD[i].partitionSectorsPerCluster * uSD[i].partitionCountOfClusters
}
      if(uSDCogChecker[0])
        totalSectors[0] := uSDMas.partitionSectorsPerCluster * uSDMas.partitionCountOfClusters
      if(uSDCogChecker[1])
        totalSectors[1] := uSDCli[0].partitionSectorsPerCluster * uSDCli[0].partitionCountOfClusters
      if(uSDCogChecker[2])
        totalSectors[2] := uSDCli[1].partitionSectorsPerCluster * uSDCli[1].partitionCountOfClusters
      if(uSDCogChecker[3])
        totalSectors[3] := uSDCli[2].partitionSectorsPerCluster * uSDCli[2].partitionCountOfClusters




            
      
      'Place this Error as final check before proceeding to copy
      if(errorSectorMismatch)
        'Notify Error to user

        Stop(statusAddress)
        


      else
        '!!! !!! !!! !!! !!
        '!!! Good-To-Go !!!
        '!!! !!! !!! !!! !!
        'Proceed with copying      
        '!!! !!! !!! !!! !!

        '***
        '*** Debugging User Only
        '***
        long[statusAddress] |= _debugFlag2   'Set Bit  - PASSED
        repeat i from 0 to 3
          'Displaying which uSD Socket is committed
          DBG.Str(String(13,"uSD Socket ["))
          DBG.Dec(i)
          DBG.Str(String("] is : "))
          DBG.Dec(uSDCogChecker[i])
          
        repeat i from 0 to 3
          'Displaying which uSD Socket is committed
          DBG.Str(String(13,"Total Sectors for Card ["))
          DBG.Dec(i)
          DBG.Str(String("] is : "))
          DBG.Dec(totalSectors[i])
        '***
        


        long[statusAddress] |= _OpsOrNot  'Turn In Operation Flag ON  
        curSector := 0
        repeat while not (long[statusAddress] & _UserInt)  'While user did not press STOP button
          '***
          '*** Debugging Only
          '***
          DBG.Str(String(13, "Inside Copying Loop ..."))
          '***

        
        
          if(curSector =< totalSectors[0]) 

            '***
            '*** Debugging Only
            '***
            DBG.Str(String(13, "Inside curSector =< totalSector ..."))
            DBG.Str(String(13, "Before readLowLevelSector from Master ..."))
            DBG.Str(String(13,13, "Master Drive"))
            DBG.Str(String(13, "============"))
            DBG.Str(String(13, "Disk Signature : "))
            DBG.Dec(uSDMas.partitionDiskSignature)
            DBG.Str(String(13, "Volume ID : "))
            DBG.Dec(uSDMas.partitionVolumeIdentification)
            DBG.Str(String(13, "Parition Type: "))
            DBG.Str(uSDMas.partitionFileSystemType)
            DBG.Str(String(13, "Volume Label : "))
            DBG.Str(uSDMas.partitionVolumeLabel)
            DBG.Str(String(13, "Total Sectors : "))
            DBG.Dec(uSDMas.partitionSectorsPerCluster * uSDMas.partitionCountOfClusters)
            
            DBG.Str(String(13,13, "Client Drive 1"))
            DBG.Str(String(13, "=============="))
            DBG.Str(String(13, "Disk Signature : "))
            DBG.Dec(uSDCli[0].partitionDiskSignature)
            DBG.Str(String(13, "Volume ID : "))
            DBG.Dec(uSDCli[0].partitionVolumeIdentification)
            DBG.Str(String(13, "Parition Type: "))
            DBG.Str(uSDCli[0].partitionFileSystemType)
            DBG.Str(String(13, "Volume Label : "))
            DBG.Str(uSDCli[0].partitionVolumeLabel)
            DBG.Str(String(13, "Total Sectors : "))
            DBG.Dec(uSDcli[0].partitionSectorsPerCluster * uSDCli[0].partitionCountOfClusters)

            DBG.Str(String(13, 13, "curSector Value is : "))
            DBG.Dec(curSector)
            '***



            'Read Master Sector
            
            result := \uSDMas.readLowLevelSector(curSector, @readSectorBlockBuffer)
            Pause(_waitToFlush)  'Give it a little time
            'Pause(2000)



            '***
            '*** Debugging Only
            '***
            DBG.Str(String(13, "Debugging: "))
            DBG.Str(result)            
            DBG.Str(String(13, "After readLowLevelSector of Master ... readSectorBlockBuffer:", 13))
            repeat i from 0 to 127
              DBG.Hex(long[@readSectorBlockBuffer][i], 8)
              DBG.Str(String(","))
            repeat
            '***
            
{
            'Loop to Write to Clients
            repeat i from 1 to 3
              if(uSDCogChecker[i])
                if(curSector =< totalSectors[i])  'Check to make sure not writing to some phantom sector
                  uSD[i].writeLowLevelSector(curSector, @writeSectorBlockBuffer)
                  Pause(_waitToFlush)

                '***
                '*** Debugging Only
                '***
                DBG.Str(String(13, "Written Sector: "))
                DBG.Dec(curSector)
                Pause(50)
                '***
 }                

            curSector++

          else
            'Sector-Copy from Master to Client successful!
            '1. Flag for success
            long[statusAddress] |= _opsSuccess

            '2. Quit
            quit



             
        long[statusAddress] ^= long[statusAddress] & _OpsOrNot  'Turn In Operation Flag OFF  

             
     
     
     
    else
      '@@@ @@@ @@@ @@@ @@@
      '@@@  FILE COPY  @@@  
      '@@@ @@@ @@@ @@@ @@@

     


  



  
  

  return    
  


PUB Stop(statusAddress)
{{
  Unloading this already loaded
}}

  if( (long[statusAddress] & _masCrdDet) and uSDCogChecker[0] )

    uSDMas.FATEngineStop
    uSDCogChecker[0] := 0

  if( (long[statusAddress] & _c1CrdDet) and uSDCogChecker[1] )
    'Client 1
    uSDCli[0].FATEngineStop
    uSDCogChecker[1] := 0

  if( (long[statusAddress] & _c2CrdDet) and uSDCogChecker[2] )
    'Client 2
    uSDCli[1].FATEngineStop
    uSDCogChecker[2] := 0

  if( (long[statusAddress] & _c3CrdDet) and uSDCogChecker[3] )
    'Client 3
    uSDCli[2].FATEngineStop
    uSDCogChecker[3] := 0

  return


PRI Pause(ms) | t
{{Delay program ms milliseconds}}

  t := cnt - 1088                                               ' sync with system counter
  repeat (ms #> 0)                                              ' delay must be > 0
    waitcnt(t += msecValue)
  